//+------------------------------------------------------------------+
//|                                                  SafeNumbers.mqh |
//|                                 Copyright  2011-2016, Marketeer |
//|                          https://www.mql5.com/en/users/marketeer |
//| MQL's OOP notes: rubber arrays, safe numbers,                    |
//|                     exceptions handling and other tricks: part 1 |
//|                        https://www.mql5.com/en/blogs/post/683241 |
//+------------------------------------------------------------------+

template<typename T>
class Number
{
  private:
    T v;
    double eps;
    void init(const double e)
    {
      eps = e == 0 ? MathPow(10, -1 * _Digits) : e;
    }
    
  public:
    Number() {init(0);}
    Number(const T &d, const double e = 0): v(d) {init(e);}
    Number(T d, const double e = 0): v(d) {init(e);}
    Number(const Number<T> &c): v(c.v) {init(c.eps);}
    
    T operator=(T x)
    {
      v = x;
      return v;
    }
    Number<T> operator=(const Number<T> &x)
    {
      v = x.v;
      return this;
    }
    bool operator==(T x) const
    {
      return MathAbs(v - x) < eps;
    }
    bool operator!=(T x) const
    {
      return !(this == x);
    }
    
    T operator+(const Number<T> &x) const
    {
      return v + x.v;
    }
    T operator-(const Number<T> &x) const
    {
      return v - x.v;
    }
    T operator/(Number<T> &x) const
    {
      if(x.v == 0)
      {
        #ifdef EXCEPTIONS
          return (T)try::throw("Divide by zero");
        #else
          return 0;
        #endif
      }
      return v / x.v;
    }
    T operator*(Number<T> &x) const
    {
      return v * x.v;
    }
    
    bool operator<(Number<T> &x) const
    {
      return v < x.v;
    }
    bool operator>(Number<T> &x) const
    {
      return v > x.v;
    }
    bool operator<=(Number<T> &x) const
    {
      return v <= x.v;
    }
    bool operator>=(Number<T> &x) const
    {
      return v >= x.v;
    }
    
    T operator+=(const Number<T> &x)
    {
      return v += x.v;
    }
    T operator-=(const Number<T> &x)
    {
      return v -= x.v;
    }
    T operator*=(const Number<T> &x)
    {
      return v *= x.v;
    }
    T operator/=(const Number<T> &x)
    {
      if(x.v == 0)
      {
        v = 0;
        #ifdef EXCEPTIONS
          return (T)try::throw("Divide by zero");
        #else
          return 0;
        #endif
      }
      return v /= x.v;
    }
    T operator++()
    {
      return ++v;
    }
    T operator--()
    {
      return --v;
    }
    
    // this is required because MQL does not support type casting operators overloading
    string toString()
    {
      return (string)v;
    }
};
